/*
 * probe.c: helping you probe things
 *
 * this file is responsible for figuring out the
 * continuity and diode tests from the raw analog data.
 * 
 * todo: confirm mathematically that continuity values
 *       match the measured values
 *
 * file creation: 20231016 0255
 */

#include "py32f0xx_conf.h"

#include "testo.h"

#include "adc.h"
#include "led.h"
#include "userio.h"



#define CONT_OPEN			1240



// diode measurements at 3V3
#define DIODE_OPEN			3860		// test unit measures ~4060 no D1, ~3960 with D1
#define DIODE_SCHOTTKY		1960		// vishay 40V 1A SB140
#define DIODE_STANDARD		2330		// fairchild 5V2 zener, forward biased
#define DIODE_LED_BLUE		3190
#define DIODE_LED_GRN		3040		// very dim
#define DIODE_LED_RED		3020
#define DIODE_SHORT			1900		// test unit measures ~1820


#define CONT_SHORTED		858			// worst case measured value of short
#define CONT_47OHM			869			// measured value of 47ohm 1% resistor at room temp
#define CONT_100OHM			886			// measured value of 100ohm 1% resistor at room temp
#define CONT_EXTRA_COUNTS	8			// extra raw counts as a fudge factor, post-scaling
#define CONT_LATCH_THRESH	15			// extra raw counts for latching indicator, post-scaling

#define CONT_LATCH_TIMEOUT	4			// probe measurement passes to hold latched indicator



static uint16_t vref;

static uint16_t cont_thresh_set = 0;
static const uint16_t cont_thresh[] = {
	CONT_100OHM, CONT_47OHM
};

static uint8_t latch = 0;
static uint8_t buzzer = 0;



void probe_set_thresh(uint8_t set)
{
	cont_thresh_set = set;
}

uint8_t probe_get_thresh()
{
	return cont_thresh_set;
}

static inline void probe_measure_cont()
{
	uint16_t probe, v_ext;
	uint32_t x, c;
	uint8_t g, b;

	probe = adc_avg[ADC_PROBE];
	v_ext = adc_avg[ADC_VREF_EXT];

	// if the button has been pushed, toggle the buzzer
	if (userio_get_btn() > 0) {
		buzzer ^= 1;
	}

	// if the button has been held, change the threshold
	if (userio_get_btn_held() == 60) {
		userio_set_btn_override();
		cont_thresh_set ^= 1;
	}

	// this LED will not set anything,
	// but zero it out anyway
	led_setrgb(1, 0, 0, 0);
	
	if (vref) {
		// probe level
		x  = probe << 12;					// maximum possible level
		x /= vref;							// normalize to 4096max
		if (x > 4095) x = 4095;

		// threshold level
		c  = cont_thresh[cont_thresh_set] << 12;
		c /= vref;

		// are we measuring a lower value?
		if (x < (c + CONT_EXTRA_COUNTS)) {	// roughly 100ohm or lower
			latch = CONT_LATCH_TIMEOUT;		// latch any continuity for a while
		}

		b = buzzer ? 100 : 0;

		if (latch) {
			// hysteresis
			if (x >= c + CONT_LATCH_THRESH) latch--;

			// indicate continuity
			led_setrgb(0, 0, 250, b);
			led_buzz(0);
			if (buzzer) led_buzz(1);
		} else {
			// idle
			g = cont_thresh_set ?  40 :  0;
			led_setrgb(0, 120, g, b);
			led_buzz(0);
		}
	}

	if (!vref) vref = v_ext;
	vref += v_ext;
	vref >>= 1;
}

static inline void probe_measure_diode()
{
	uint16_t p;

	p = adc_avg[ADC_PROBE];

	led_setrgb(0, 0, 0, 0);

	if (p < DIODE_SHORT) {
		// show off on short
		led_setrgb(1, 0, 0, 0);
	} else
	if (p >= DIODE_OPEN) {
		// show red on open
		led_setrgb(1, 120, 0, 0);
	} else
	if (p >= 1940 && p < 2650) {
		// show green on regular or schottky diodes
		led_setrgb(1, 0, 200, 0);
	} else 
	if (p >= 2650 && p < 3700) {
		// show blue on LEDs
		led_setrgb(1, 0, 0, 500);
	} else
	if (p >= 3701) {
		// show red+blue if between open and LED
		led_setrgb(1, 120, 0, 120);
	}
}

void probe_measure()
{
	uint8_t mode;

	mode = userio_get_mode();

	switch (mode) {
		case MODE_CONT:		probe_measure_cont(); break;
		case MODE_FUN:		break;	// todo: run RGBLED program
		case MODE_DIODE:	probe_measure_diode(); break;
	}
}

/*
 * switch probe mode.
 * sets up LED peripheral and mode pin as necessary.
 */
void probe_mode_switch(uint8_t mode)
{
	if (mode == MODE_CONT) {
		// set LED to buzzer mode, set PROBESEL low
		led_mode_buzzer();
		PROBESEL_PORT->BRR  = (1 << PROBESEL_PIN);
	} else {
		// set LED to dual RGB mode, set PROBESEL high
		led_mode_rgb();
		PROBESEL_PORT->BSRR = (1 << PROBESEL_PIN);
	}
}